import 'package:flutter/material.dart';
import 'package:presentation/src/page_transformer.dart';
import 'package:presentation/src/presentation_controller.dart';

class Presentation extends StatelessWidget {
  const Presentation({
    required this.children,
    required this.controller,
    required this.presentationController,
    this.enableMouseNavigation = true,
    super.key,
  });

  /// Allows navigating to next slide with single tap and to previous slide
  /// with double tap
  final bool enableMouseNavigation;

  final List<Widget> children;
  final PageController controller;
  final PresentationController presentationController;

  @override
  Widget build(BuildContext context) {
    return NestedScrollView(
      headerSliverBuilder: (BuildContext context, bool innerBoxIsScrolled) {
        return [
          SliverOverlapAbsorber(
            handle: NestedScrollView.sliverOverlapAbsorberHandleFor(context),
          ),
        ];
      },
      body: GestureDetector(
        key: const Key('presentation'),
        onTap: enableMouseNavigation ? presentationController.nextStep : null,
        onDoubleTap:
            enableMouseNavigation ? presentationController.previousStep : null,
        child: ScrollNotifier(
          child: PageView.builder(
            controller: controller,
            itemCount: children.length,
            itemBuilder: (context, index) {
              return PageViewSettings(
                index: index,
                child: children[index],
              );
            },
          ),
        ),
      ),
    );
  }
}

class PageViewSettings extends InheritedWidget {
  const PageViewSettings({
    super.key,
    required this.index,
    required super.child,
  });

  final int index;

  static PageViewSettings of(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType()!;

  @override
  bool updateShouldNotify(PageViewSettings oldWidget) =>
      index != oldWidget.index;
}

/// Based on https://github.com/roughike/page-transformer
/// A class that can be used to compute visibility information about
/// the current page.
class PageVisibilityResolver {
  const PageVisibilityResolver({
    ScrollMetrics? metrics,
    double? viewPortFraction,
  })  : _pageMetrics = metrics,
        _viewPortFraction = viewPortFraction;

  final ScrollMetrics? _pageMetrics;
  final double? _viewPortFraction;

  /// Calculates visibility information for the page at [pageIndex].
  /// Used inside PageViews' itemBuilder, but can be also used in a
  /// simple PageView that simply has an array of children passed to it.
  PageVisibility resolvePageVisibility(int pageIndex) {
    final pagePosition = _calculatePagePosition(pageIndex);
    final visiblePageFraction =
        _calculateVisiblePageFraction(pageIndex, pagePosition);

    return PageVisibility(
      visibleFraction: visiblePageFraction,
      pagePosition: pagePosition,
    );
  }

  double _calculateVisiblePageFraction(int index, double pagePosition) =>
      pagePosition > -1.0 && pagePosition <= 1.0
          ? 1.0 - pagePosition.abs()
          : 0.0;

  double _calculatePagePosition(int index) {
    final viewPortFraction = _viewPortFraction ?? 1.0;
    final pageViewWidth =
        (_pageMetrics?.viewportDimension ?? 1.0) * viewPortFraction;
    final pageX = pageViewWidth * index;
    final scrollX = _pageMetrics?.pixels ?? 0.0;
    final pagePosition = (pageX - scrollX) / pageViewWidth;
    final safePagePosition = !pagePosition.isNaN ? pagePosition : 0.0;

    if (safePagePosition > 1.0) {
      return 1;
    } else if (safePagePosition < -1.0) {
      return -1;
    }

    return safePagePosition;
  }
}

/// A class that contains visibility information about the current page.
class PageVisibility {
  const PageVisibility({
    required this.visibleFraction,
    required this.pagePosition,
  });

  /// How much of the page is currently visible, between 0.0 and 1.0.
  ///
  /// For example, if only the half of the page is visible, the
  /// value of this is going to be 0.5.
  ///
  /// This doesn't contain information about where the page is
  /// disappearing to or appearing from. For that, see [pagePosition].
  final double visibleFraction;

  /// Tells the position of this page, relative to being visible in
  /// a "non-dragging" position, between -1.0 and 1.0.
  ///
  /// For example, if the page is fully visible, this value equals 0.0.
  ///
  /// If the page is fully out of view on the right, this value is
  /// going to be 1.0.
  ///
  /// Likewise, if the page is fully out of view, on the left, this
  /// value is going to be -1.0.
  final double pagePosition;
}
